%% tclp__cgi.pl main file for running the cgi version of tclp
%% Copyright (C) 2001,2002 INRIA Emmanuel Coquery

%% This library is free software; you can redistribute it and/or
%% modify it under the terms of the GNU Lesser General Public
%% License as published by the Free Software Foundation; either
%% version 2.1 of the License, or (at your option) any later version.

%% This library is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
%% Lesser General Public License for more details.

%% You should have received a copy of the GNU Lesser General Public
%% License along with this library; if not, write to the Free Software
%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%                                                                           %%
%%                                  TCLP                                     %%
%%                                                                           %%
%%         Emmanuel Coquery, projet Contraintes, INRIA Rocquencourt          %%
%%                                                                           %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% $Header: /home/beaune/contraintes/coquery/cvs/contraintes/tclpchr/src/tclp__cgi.pl,v 1.5 2002/09/19 15:36:00 coquery Exp $

:- use_module(library('tclp/tclp__utils')).
:- use_module(library('tclp/tclp__arrays')).
:- use_module(library('tclp/tclp__def')).
:- use_module(library('tclp/tclp__arity')).
:- use_module(library('tclp/tclp__reader')).
:- use_module(library('tclp/tclp__errors')).
:- use_module(library('tclp/tclp__type_order')).
:- use_module(library('tclp/tclp__clp_types')).
:- use_module(library('tclp/tclp__handle_goals')).
:- use_module(library('tclp/tclp__declarations')).
:- use_module(library('tclp/tclp__files')).
:- use_module(library('tclp/tclp__functor_types')).
:- use_module(library('tclp/tclp__type_checker')).
:- use_module(library('tclp/tclp__connexity')).
:- use_module(library('tclp/tclp__type_inference')).
:- use_module(library('tclp/sicstus_modules')).

:- use_module(library(ordsets)).
:- use_module(library(lists)).
:- use_module(library(pillow)).
:- use_module(library(charsio)).

%% Verbose mode : debug_main actions

%:- enable_verbose(debug_main).
%:- enable_verbose(debug_handle_goals).
%:- enable_verbose(debug_functor_types).
%:- enable_verbose(debug_files).
%:- enable_verbose(actions).
%:- enable_verbose(debug_cgi,true).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Process a file
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- include('version.pl').

%% enable/disable type inference
:- dynamic do_inference/0.

set_inference(DoIt) :-
        (
          DoIt
        ->
          asserta(do_inference)
        ;
          retractall(do_inference)
        ).

print_lines([]).
print_lines([L|Rest]) :-
        print_line(L), print_lines(Rest).

print_line([]) :- nl.
print_line([C|Rest]) :-
        put_code(C), print_line(Rest).

%% Type checks one connex part
process_one_connex_part(ConnexPart) :-
        verbose(debug_main,'starting connex part ...'),
        (
          do_inference
        ->
          verbose(debug_main, 'inference'),
          infer_connex_part(ConnexPart)
        ;
          (
            verbose(debug_main, 'type check'),
            type_check_clauses(ConnexPart),!
          ;
            print_error('type check failed')
          )
        ),
        verbose(debug_main,'connex part ok').
process_one_connex_part(_) :-
        print_error('Connex part failed').

%% Type check all connex parts
process_connex_parts([]).
process_connex_parts([ Part | Next ]) :-
        process_one_connex_part(Part),
        verbose(debug_main,['\nProcessing :\n',Part,'\nOK\n']),
        process_connex_parts(Next),
        verbose(debug_main,['\nProcessing :\n',Part]).

%% Process the type checking of a list of lines
process_lines(Name, LineList) :-
        chrono('declarations handling',
               (
                 read_from_lines(Name, LineList, PhraseList),
                 verbose(debug_main,'read file ok'),
                 build_ordered_connex_part_list(PhraseList, OrderedPhrases),
                 build_order
               )
              ),
        chrono('type cheking',process_connex_parts(OrderedPhrases)).

%% Process the type checking of a list of phrases
process_phrases(PhraseList) :-
        chrono('declarations handling',
               (
                 build_ordered_connex_part_list(PhraseList, OrderedPhrases),
                 build_order
               )
              ),
        chrono('type cheking',process_connex_parts(OrderedPhrases)).

:- dynamic tclp_home/1.
tclp_home('.').

%% parse the command line
parse_args([]) :- !.
parse_args([Ignore|Rest]) :- (Ignore=en;Ignore=fr),!, parse_args(Rest).
parse_args(['--home',Home|Rest]) :- !,
        asserta(tclp_home(Home)),
        parse_args(Rest).
parse_args(Args) :-
        write('! unknown args : '), write(Args),nl.

%% Proceed to the reading of the standard type library
read_stdlib(PrologEngine, CustomStdLibName, CustomStdLibData) :-
        (
          PrologEngine == 'customstdlib'
        ->
          load_from_lines(CustomStdLibName,CustomStdLibData)
        ;
          tclp_home(TCLP_Home),
          (
            atom_concat(_,'/',TCLP_Home)
          ->
            atom_concat(TCLP_Home,'lib/',LibHome)
          ;
            atom_concat(TCLP_Home,'/lib/',LibHome)
          ),
          (
            PrologEngine == 'iso'
          ->
            atom_concat(LibHome,'ISO/',StdLibDir)
          ;
            PrologEngine == 'gnu'
          ->
            atom_concat(LibHome,'gprolog/',StdLibDir)
          ;
            atom_concat(LibHome,'sicstus/', StdLibDir),
            enable_sicstus
          ),
          add_include(StdLibDir),
          load_file('stdlib.typ')
        ).

%% loads a file or a text area
load_from_area_or_file(Name,AreaData,FileName,FileData) :-
        (
          form_empty_value(FileName)
        ->
          load_from_lines(Name,AreaData)
        ;
          load_from_lines(FileName,FileData)
        ).

%% reads a file or a text area
read_from_area_or_file(Name,AreaData,FileName,FileData,Phrases) :-
        (
          form_empty_value(FileName)
        ->
          read_from_lines(Name,AreaData,Phrases)
        ;
          read_from_lines(FileName,FileData,Phrases)
        ).

%% type checks the phrases
run_type_check(PhraseList) :-
        process_phrases(PhraseList).

%% set the proper options
set_options :-
        set_label_functors(true),
        enable_verbose(types,false),
        enable_verbose(defaults,true),
        set_inference(true).

%% print the type of the argument
print_type(Input) :-
        (
          form_empty_value(Input)
        ->
          print_error('No functor description found !!!')
        ;
          true
        ),
        atom_concat(Input,' .',AtomToParse),
        atom_codes(AtomToParse, CodesToParse),
        open_chars_stream(CodesToParse, Stream),
        read_one_term(Stream, Term, _, 'functor description'),
        close(Stream),
        (
          Term = F/N, atom(F), (integer(N);var(N))
        ->
          true
        ;
          functor(Term,F,N)
        ),
        print_functor_types(F/N).

%% prints a module file
print_module(RawModuleName) :-
        remove_spaces(RawModuleName, ModuleName),
        (
          (form_empty_value(ModuleName) ; ModuleName = '')
        ->
          print_error('No module name found !!!')
        ;
          true
        ),
        (
          atom_concat(_,'.typ',ModuleName)
        ->
          FileName = ModuleName
        ;
          atom_concat(ModuleName,'.typ',FileName)
        ),
        print_file(FileName).
          

%% file form handler
get_file_form_value(Dict,Name,FileName,Data) :-
        get_form_value(Dict,Name,A),
        (
          form_empty_value(A)
        ->
          form_empty_value(FileName),
          form_empty_value(Data)
        ;
          A=file(FileName,Data)
        ).

%% Handles the form data
handle_form_data :-
        verbose(debug_cgi,'command line args'),
        %% parse arguments
        prolog_flag(argv, Args),
        parse_args(Args),

        verbose(debug_cgi,'default options'),
        %% sets the default tclp options
        set_options,

        verbose(debug_cgi,'parse input'),
        %% parse form input
        get_form_input(Dict),
        verbose(debug_cgi,'get values'),
        %% Corresponds to the submit button for type checking
        get_form_value(Dict,typesubmit,TypeSubmitButton),
        %% Corresponds to the submit button for type checking
        get_form_value(Dict,checksubmit,CheckSubmitButton),
        %% source of the type checked program
        get_file_form_value(Dict,sourcefiledata,
                            SourceFileName, SourceFileData),
        get_form_value(Dict,source,SourceLineList),
        %% types for the type checked program
        get_file_form_value(Dict,typesfiledata,
                            TypesFileName, TypesFileData),
        get_form_value(Dict,types,TypesLineList),
        %% the submit button for type interrogation
        get_form_value(Dict,intersubmit,WhichTypeButton),
        %% field containing the type of which the user wants infos
        get_form_value(Dict,question,WhichTypeValue),
        %% the submit button for module interrogation
        get_form_value(Dict,moduleview,WhichModuleButton),
        %% field containing the name of the module of which the user
        %% wants infos
        get_form_value(Dict,modname,WhichModuleValue),
        %% which prolog engine is used
        get_form_value(Dict,prolog_engine,PrologEngine),
        %% file data corresponding to a custom stdlib
        get_file_form_value(Dict,customstdlibdata,
                            CustomStdLibName,CustomStdLibData),
        %% submit for viewing stdlib
        get_form_value(Dict,stdlibsubmit,StdLibSubmitButton),

        verbose(debug_cgi,'load stdlib'),
        %% load the stdlib
        read_stdlib(PrologEngine, CustomStdLibName, CustomStdLibData),
        
        verbose(debug_cgi,'read declarations'),
        %% read declarations
        load_from_area_or_file(types,TypesLineList,
                               TypesFileName,TypesFileData),
        read_from_area_or_file(source,SourceLineList,
                               SourceFileName, SourceFileData,
                               PhraseList),

        %% choose what action to take w.r.t the submit button that was clicked

        (
          StdLibSubmitButton \= ''
        ->
          write('%% printing the standard library ...'),nl,
          (
            PrologEngine == 'customstdlib'
          ->
            print_lines(CustomStdLibData)
          ;
            print_module(stdlib)
          )
        ;
          TypeSubmitButton \= ''
        ->
          set_inference(true),
          write('%% starting type checking ...'),nl,
          %% type checking
          run_type_check(PhraseList)
        ;
          CheckSubmitButton \= ''
        ->
          set_inference(false),
          write('%% starting type checking ...'),nl,
          %% type checking
          run_type_check(PhraseList)
        ;
          WhichTypeButton \= ''
        ->
          write('%% starting type interrogation ...'),nl,
          print_type(WhichTypeValue)
        ;
          WhichModuleButton \= ''
        ->
          write('%% starting module interrogation ...'),nl,
          print_module(WhichModuleValue)
        ;
          write('! Unknown action : stopping'),nl
        ),
        write('%% finished'),nl
        ;
        write('! failed'),nl,
        true. % (to avoid internal http error)

run :-
        current_prolog_flag(user_output,UserOut),
        set_prolog_flag(user_error,UserOut),

        %% output the header
        write('Content-type: text/plain'),nl,nl,

        tclp__version(V),
        verbose(always,['tclp ',V]),
        catch(handle_form_data,E,print_exception(E));true.

user:runtime_entry(start) :- run.
